from __future__ import division, print_function
import os
import sys
import pandas
import logging
from VAPr.annovar import AnnovarWrapper
from VAPr.parsers import VariantParsing
from VAPr.ingester import Ingester
from pymongo import MongoClient

logger = logging.getLogger()
logger.setLevel(logging.INFO)
try:
    logger.handlers[0].stream = sys.stdout
except:
    pass


__author__ = 'Carlo Mazzaferro<cmazzafe@ucsd.edu>'


class ProjectData(object):
    """ Base class for the project instantiation classes that will run the necessary functions. Mostly a container. """

    supported_build_vers = ['hg19', 'hg18', 'hg38']

    def __init__(self,
                 input_dir,
                 output_dir,
                 annovar_path,
                 project_data,
                 split_vcf=False,
                 design_file=None,
                 build_ver=None,
                 mongod_cmd=None):

        self.input_dir = input_dir
        self.output_csv_path = output_dir
        self.design_file = design_file
        self.annovar = annovar_path
        self.project_data = project_data
        self.buildver = self.check_ver(build_ver)
        self.times_called = 0
        self.mapping = self.get_mapping()
        self.split = split_vcf
        self.mongod = mongod_cmd

    def get_mapping(self):
        """
        Get mapping of vcf file

        Each vcf file has a dictionary of values associated with it of the following form:

             mapping = {'raw_vcf_file_full_path': full path of vcf file,
                        'vcf_file_basename': vcf file base name,
                        'csv_file_basename': csv file base name,
                        'sample_names': samples in vcf file,
                        'num_samples_in_csv': number of samples,
                        'csv_file_full_path': output of sample file where csv files will be stored,
                        'vcf_sample_dir': directory of sample where vcf file lives,
                        'extra_data': None or some dictionary of data contained in design file
                        }

        Each vcf file will be moved to a directory named after the samples it contains

        """
        if self.design_file:
            design_df = pandas.read_csv(self.design_file)
            sample_to_vcf_map = self.check_file_existance_return_mapping(design_df)

        else:
            sample_to_vcf_map = self.check_file_existance_return_mapping_no_design_file()

        return sample_to_vcf_map

    @staticmethod
    def listdir_fullpath(d):
        """ Helper function to list full path of files in directory """
        return [os.path.join(d, f) for f in os.listdir(d)]

    def check_file_existance_return_mapping_no_design_file(self):
        """ Ingest all files in specified directory and return mapping """

        organizer = Ingester(self.input_dir, self.output_csv_path)
        organizer.walk()
        return organizer.mapping_list

    def check_file_existance_return_mapping(self, design_df):
        """ Ingest design file and the directories/files referenced in it and return mapping """

        organizer = Ingester(self.input_dir, self.output_csv_path)
        organizer.digest_design_file(design_df)
        return organizer.mapping_list

    def check_ver(self, build_ver):
        """ Make sure genome version is acceptable """

        if not build_ver:
            self.buildver = 'hg19'  # Default genome build version

        if build_ver not in self.supported_build_vers:
            raise ValueError('Build version must not recognized. Supported builds are'
                             ' %s, %s, %s' % (self.supported_build_vers[0],
                                              self.supported_build_vers[1],
                                              self.supported_build_vers[2]))
        else:
            self.buildver = build_ver

        return self.buildver


class AnnotationProject(ProjectData):

    def __init__(self,
                 input_dir,
                 output_dir,
                 annovar_path,
                 project_data,
                 design_file=None,
                 build_ver=None,
                 mongod_cmd=None):

        """ Class that implements the API and the major annotation/saving methods  """

        super(AnnotationProject, self).__init__(input_dir,
                                                output_dir,
                                                annovar_path,
                                                project_data,
                                                design_file=design_file,
                                                build_ver=build_ver,
                                                mongod_cmd=mongod_cmd)

        self.annovar_wrapper = AnnovarWrapper(self.input_dir,
                                              self.output_csv_path,
                                              self.annovar,
                                              self.project_data,
                                              self.mapping,
                                              design_file=self.design_file,
                                              build_ver=self.buildver)

        self.annotator_wrapper = VariantParsing(self.input_dir,
                                                self.output_csv_path,
                                                self.annovar,
                                                self.project_data,
                                                self.mapping,
                                                design_file=self.design_file,
                                                build_ver=self.buildver,
                                                mongod_cmd=self.mongod)

    def download_dbs(self, all_dbs=True, dbs=None):
        """ Wrapper around Annovar database downloading function """
        self.annovar_wrapper.download_dbs(all_dbs=all_dbs, dbs=dbs)

    def run_annovar(self, batch_jobs=10, multisample=False):
        """ Wrapper around multiprocess Annovar annotation  """
        self.annovar_wrapper.run_annovar(batch_jobs=batch_jobs, multisample=multisample)

    def annotate_and_save(self, buffer_vars=False, verbose=2):
        """ Wrapper around annotation runner. Deprecated in favour of parallel parsing """
        self.annotator_wrapper.annotate_and_saving(buffer_vars=buffer_vars, verbose=verbose)

    def parallel_annotation_and_saving(self, n_processes=4, verbose=1):
        """ Wrapper around parallel annotation multiprocess runner  """
        self.annotator_wrapper.parallel_annotation(n_processes=n_processes, verbose=verbose)

    def quick_annotate_and_save(self, n_processes=8):
        """ Wrapper around parallel annotation multiprocess runner using MyVariant solely """
        self.annotator_wrapper.quick_annotate_and_save(n_processes=n_processes)

    def write_output_files_by_sample(self):
        """ Wrapper around function that implemts the writing of csv files for each sample in collection """
        self.annotator_wrapper.generate_output_files_by_sample()




import myvariant
import vcf
import csv
import re
import itertools
import logging
import sys
import VAPr.vcf_parsing as vvp
import VAPr.definitions as definitions
logger = logging.getLogger()
logger.setLevel(logging.INFO)
try:
    logger.handlers[0].stream = sys.stdout
except:
    pass


class HgvsParser(object):

    """ Class that process vcf files and extracts their hgvs ids"""

    def __init__(self, vcf_file):

        self.vcf = vcf_file
        # self.num_lines = sum(1 for _ in open(self.vcf))
        self.chunksize = definitions.chunk_size
        self.samples = vcf.Reader(open(self.vcf, 'r')).samples
        self.num_samples = len(self.samples)

    def get_num_lines(self):
        return sum(1 for _ in open(self.vcf))

    def get_all_variants_from_vcf(self):
        """ For debugging mostly """

        list_ids = []

        reader = vcf.Reader(open(self.vcf, 'r'))
        for record in reader:
            list_ids.append(myvariant.format_hgvs(record.CHROM, record.POS,
                                                  record.REF, str(record.ALT[0])))

        return self.complete_chromosome(list_ids)

    def get_variants_from_vcf(self, step):
        """
        Retrieves variant names from a LARGE vcf file.
        :param step: tells the parallel processing where to start and end the hgvs id creation
        :return: a list of variants formatted according to HGVS standards
        """
        reader = vcf.Reader(open(self.vcf, 'r'))
        list_ids = []

        for record in itertools.islice(reader, step * self.chunksize, (step + 1) * self.chunksize):
            list_ids.append(myvariant.format_hgvs(record.CHROM, record.POS,
                                                  record.REF, str(record.ALT[0])))

        return self.complete_chromosome(list_ids)

    @staticmethod
    def complete_chromosome(expanded_list):
        """ Ensuring syntax consistency """

        for i in range(0, len(expanded_list)):
            if 'M' in expanded_list[i]:
                one = expanded_list[i].split(':')[0]
                two = expanded_list[i].split(':')[1]
                if 'MT' not in one:
                    one = 'chrMT'
                expanded_list[i] = "".join([one, ':', two])
        return expanded_list


class TxtParser(object):

    """ Class that process an Annovar created csv file """

    def __init__(self, txt_file, samples=None, extra_data=None):

        self.samples = samples
        self.txt_file = txt_file
        self.num_lines = sum(1 for _ in open(self.txt_file))
        self.chunksize = definitions.chunk_size
        self.offset = 0
        self.extra_data = extra_data
        self.hg_19_columns = ['chr',
                              'start',
                              'end',
                              'ref',
                              'alt',
                              'func_knowngene',
                              'gene_knowngene',
                              'genedetail_knowngene',
                              'exonicfunc_knowngene',
                              'tfbsconssites',
                              'cytoband',
                              'genomicsuperdups',
                              '1000g2015aug_all',
                              'esp6500siv2_all',
                              'cosmic70',
                              'nci60',
                              'otherinfo']

        self.hg_18_columns = self.hg_19_columns

        self.hg_38_columns = ['chr',
                              'start',
                              'end',
                              'ref',
                              'alt',
                              'func_knowngene',
                              'gene_knowngene',
                              'genedetail_knowngene',
                              'exonicfunc_knowngene',
                              'cytoband',
                              'genomicsuperdups',
                              '1000g2015aug_all',
                              'esp6500siv2_all',
                              'cosmic70',
                              'nci60',
                              'otherinfo']

    def open_and_parse_chunks(self, step, build_ver=None, offset=0):

        """
        Parsing function that retrieves data from a specific location from a csv file.
        It will process every (significant) data field in an annovar-annotated csv and
        conveniently return it into a dictionary

        :param step: tells the parallel processing where to start and end the hgvs id creation
        :param build_ver: genome build version
        :param offset: deprecated, optional argument may be used later. Used to spot discrepancies between
                       csv and vcf file
        :return: list of dictionaries, where each dictionary is a variant document
        """

        listofdicts = []

        with open(self.txt_file, 'r') as txt:

            reader = csv.reader(txt, delimiter='\t')
            header = self._normalize_header(next(reader))

            for i in itertools.islice(reader, (step*self.chunksize) + self.offset,
                                      ((step+1)*self.chunksize) + offset + self.offset):

                sparse_dict = dict(zip(header[0:len(header)-1], i[0:len(header)-1]))
                sparse_dict['otherinfo'] = i[-1-len(self.samples)::]

                if build_ver == 'hg19':
                    dict_filled = {k: sparse_dict[k] for k in self.hg_19_columns if sparse_dict[k] != '.'}
                elif build_ver == 'hg18':
                    dict_filled = {k: sparse_dict[k] for k in self.hg_18_columns if sparse_dict[k] != '.'}
                else:
                    dict_filled = {k: sparse_dict[k] for k in self.hg_38_columns if sparse_dict[k] != '.'}

                modeled = AnnovarModels(dict_filled, self.samples, extra_data=self.extra_data)
                listofdicts.append(modeled.final_list_dict)

            self.offset += offset

        return listofdicts

    @staticmethod
    def _normalize_header(header):
        """ Ensuring syntax consistency """

        normalized = []

        for item in header:
            normalized.append(item.lower().replace('.', '_'))

        return normalized


class AnnovarModels(object):

    """
    The actuall class that process a single annovar-annotated variant.
    It performs specific and necessary manipulations to ensure that the data will be parsed appropriately
    and can be queried easily from MongoDB. Basically our ODM (Object Data Model) for the Annovar data.
    """

    def __init__(self, dictionary, samples, extra_data=None):

        if extra_data:
            self.dictionary = dict(extra_data.items() + dictionary.items())
        else:
            self.dictionary = dictionary
        self.samples = samples
        self.existing_keys = self.dictionary.keys()
        self.errors = None
        self.extra_data = extra_data
        self.final_list_dict = self.process()

    def process(self):
        for key in self.dictionary.keys():

            if self.dictionary['chr'] == 'chrM':
                self.dictionary['chr'] = 'chrMT'

            if key in ['1000g2015aug_all', 'esp6500siv2_all', 'nci60']:
                self.dictionary[key] = float(self.dictionary[key])

            if key in ['start', 'end']:
                self.dictionary[key] = int(self.dictionary[key])

            if key == 'cytoband':
                cytoband_data = CytoBand(self.dictionary['cytoband'])
                self.dictionary['cytoband'] = cytoband_data.fill()

            if key in ['genomicsuperdups', 'tfbsconssites']:
                self.dictionary[key] = self.to_dict(key)

            if key == 'otherinfo':
                self.dictionary[key] = [i for i in self.dictionary[key] if i != '.']

        final_annovar_list_of_dicts, self.errors = self.parse_genotype(self.dictionary)

        return final_annovar_list_of_dicts

    def parse_genotype(self, dictionary):
        """ Implements the genotype parsing scheme. Many thanks to Amanda Birmingham """

        read_depth_error = genotype_lik_error = allele_error = 0
        parser = vvp.VCFGenotypeStrings()

        list_dictionaries = []

        for index, sample in enumerate(self.samples):
            sample_specific_dict = {k: v for k, v in dictionary.items()}  # make copy, propagate genotype info over alleles

            genotype_to_fill = parser.parse(dictionary['otherinfo'][0],
                                            dictionary['otherinfo'][index + 1])

            sample_specific_dict['sample_id'] = sample
            sample_specific_dict['genotype'] = genotype_to_fill.genotype

            try:
                sample_specific_dict['filter_passing_reads_count'] = [float(genotype_to_fill.filter_passing_reads_count)]
            except ValueError:
                read_depth_error += 1
            except TypeError:
                read_depth_error += 1
            try:
                sample_specific_dict['genotype_likelihoods'] = [float(i.likelihood_neg_exponent) for i in
                                                                genotype_to_fill.genotype_likelihoods]
            except IndexError:
                genotype_lik_error += 1

            try:
                sample_specific_dict['alleles'] = [float(i.read_counts) for i in genotype_to_fill.alleles]
            except IndexError:
                allele_error += 1
            except ValueError:
                allele_error += 1

            if sample_specific_dict['genotype'] is not None:
                genotype_class_to_fill = vvp.VCFGenotypeInfo('')
                vvp.fill_genotype_class(sample_specific_dict['genotype'], genotype_class_to_fill)
                sample_specific_dict['genotype_subclass_by_class'] = genotype_class_to_fill.genotype_subclass_by_class

            list_dictionaries.append(sample_specific_dict)

        errors = (read_depth_error, genotype_lik_error, allele_error)

        return list_dictionaries, errors

    @staticmethod
    def _to_int(val):
        """ This is.... Bad """
        try:
            val = int(val)
        except:
            pass
        return val

    def to_dict(self, key):
        as_dict = dict(item.split("=") for item in self.dictionary[key].split(";"))
        as_dict["Score"] = float(as_dict["Score"])
        return as_dict


class CytoBand(object):

    """ Gets its own class cause it is particularly pesky to parse"""

    def __init__(self, cyto_band_name):

        self.letters = set('XY')
        self.name = cyto_band_name
        self.processed = self.fill()

    def fill(self):

        processed = {'Name': self.name}
        spliced = re.split('(\D+)', self.name)

        if any((c in self.letters) for c in self.name):
            processed['Chromosome'] = spliced[1][0]
            processed['Band'] = spliced[1][1]
        else:
            processed['Chromosome'] = int(spliced[0])
            processed['Band'] = spliced[1]

        processed['Region'] = spliced[2]

        if '.' in spliced:
            processed['Sub_Band'] = spliced[-1]
        return processed




"""This module exposes utility functions to validate user inputs

By convention, validation functions in this module raise an appropriate Error if validation is unsuccessful.  If it is
successful, they return either nothing or the appropriately converted input value.
"""

# personal libraries

__author__ = 'Birmingham'


def convert_to_nullable(input_val, cast_function):
    """For non-null input_val, apply cast_function and return result if successful; otherwise, return database null.

    Args:
        input_val (Any): The value to attempt to convert to either a null or the type specified by cast_function.
            The recognized null value is '.'
        cast_function (Callable[[Any], Any]): A function to cast the input_val to some specified type; should raise an
            error if this cast fails.

    Returns:
        utilities.database.NULL value if input is the null value.
        An appropriately cast value if input is not null and the cast is successful.

    Raises:
        Error: whatever error is provided by cast_function if the cast fails.
    """
    if input_val in ['.', None, '', 'NULL']:
        return -9999
    else:
        try:
            result = cast_function(input_val)
        except:
            result = -9999
    return result


def convert_to_nonneg_int(input_val, nullable=False):
    """Cast the input to a non-negative integer, if possible, or (optionally) a database null if matches null value.

    Args:
        input_val (Any): The value to attempt to convert to a non-negative integer (or a null, if nullable=True).
            The recognized null value is '.'
        nullable (Optional[bool]): True if the input value may be null, false otherwise.  Defaults to False.

    Returns:
        utilities.database.NULL value if nullable=True and the input is the null value.
        The appropriately cast non-negative integer if input is not null and the cast is successful.

    Raises:
        ValueError: if the input cannot be successfully converted to a non-negative integer (or null, if nullable=True)
    """
    err_count = 0
    if input_val == 'NULL':
        err_count += 1
        return -9999

    if nullable:
        try:
            result = convert_to_nullable(input_val, float)
        except:
            result = -9999
        if result == 'NULL':
            err_count += 1
            return -9999, err_count

    else:
        try:
            result = int(input_val)
        except TypeError:
            err_count += 1
        except ValueError:
            err_count += 1

    if not result.is_integer():
        err_count += 1
    if result < 0:
        err_count += 1
    return result, err_count



import logging
import sys
from collections import OrderedDict
import datetime

__author__ = 'Carlo Mazzaferro<cmazzafe@ucsd.edu>'

logging.basicConfig(stream=sys.stdout, level=logging.INFO)
chunk_size = 2000


# ------ #  ANNOVAR CONFIG # ------ #

down_dd = '-webfrom annovar'
annovar_hosted = OrderedDict({'knownGene': True,
                                   'tfbsConsSites': False,
                                   'cytoBand': False,
                                   'targetScanS': False,
                                   'genomicSuperDups': False,
                                   'esp6500siv2_all': True,
                                   '1000g2015aug': True,
                                   'popfreq_all_20150413': True,
                                   'clinvar_20161128': True,
                                   'cosmic70': True,
                                   'nci60': True,
                                   'avdblist': True})

dl_list_command = 'avdblist'
manual_update = {'clinvar_20161128': [datetime.datetime(2016, 11, 28)],
                 '1000g2015aug': [datetime.datetime(2016, 8, 30)],
                 'popfreq_all_20150413': [datetime.datetime(2015, 4, 13)]}

hg_18_databases = OrderedDict({'knownGene': 'g',
                                    'tfbsConsSites': 'r',
                                    'cytoBand': 'r',
                                    'targetScanS': 'r',
                                    'genomicSuperDups': 'r',
                                    'esp6500siv2_all': 'f',
                                    '1000g2015aug': 'f',
                                    'cosmic70': 'f',
                                    'nci60': 'f'})

hg_19_databases = OrderedDict({'knownGene': 'g',
                                    'tfbsConsSites': 'r',
                                    'cytoBand': 'r',
                                    'targetScanS': 'r',
                                    'genomicSuperDups': 'r',
                                    'esp6500siv2_all': 'f',
                                    '1000g2015aug': 'f',
                                    'popfreq_all_20150413': 'f',
                                    'clinvar_20161128': 'f',
                                    'cosmic70': 'f',
                                    'nci60': 'f'})

hg_38_databases = OrderedDict({'knownGene': 'g',
                                    'cytoBand': 'r',
                                    'genomicSuperDups': 'r',
                                    'esp6500siv2_all': 'f',
                                    '1000g2015aug': 'f',
                                    'clinvar_20161128': 'f',
                                    'cosmic70': 'f',
                                    'nci60': 'f'})


from vapr import base

if __name__ == '__main__':
    # Directory of input files to be annotated
    IN_PATH = "/Volumes/Carlo_HD1/CCBB/VAPr_files/vcf_benchmark/"

    # Output file directory
    OUT_PATH = "/Volumes/Carlo_HD1/CCBB/VAPr_files/csv_benchmark/"
    # Location of your annovar dowload. The folder should contain the following files/directories:
    ANNOVAR_PATH = '/Volumes/Carlo_HD1/CCBB/annovar/'

    db_name = 'VariantDBMultiBenchmark'
    collection_name = 'test_collection_small_vcf'

    # Design File (optional)
    # design_file = '/Volumes/Carlo_HD1/CCBB/VAPr_files/guorong_single_sample.csv'

    # Databse and Collection names (optional)
    proj_data = {'db_name': db_name,
                 'project_name': collection_name}

    Project = base.AnnotationProject(IN_PATH,
                                OUT_PATH,
                                ANNOVAR_PATH,
                                proj_data,
                                # design_file=design_file,
                                build_ver='hg19')

    Project.parallel_annotation_and_saving(n_processes=8, verbose=2)


from __future__ import division, print_function

import myvariant
import os
import sys
from pymongo import MongoClient
import VAPr.definitions as definitions
from VAPr.models import TxtParser, HgvsParser
from VAPr.writes import Writer
from VAPr.queries import Filters
import tqdm
from multiprocessing import Pool
import logging
import subprocess
import shlex
import time
logger = logging.getLogger()
logger.setLevel(logging.INFO)
try:
    logger.handlers[0].stream = sys.stdout
except:
    pass

__author__ = 'Carlo Mazzaferro<cmazzafe@ucsd.edu>'


class VariantParsing:

    def __init__(self,
                 input_dir,
                 output_csv_path,
                 annovar_path,
                 project_data,
                 mapping,
                 design_file=None,
                 build_ver=None,
                 mongod_cmd=None):

        """ Project data """
        self.input_dir = input_dir
        self.output_csv_path = output_csv_path
        self.annovar = annovar_path
        self.project_data = project_data
        self.design_file = design_file
        self.buildver = build_ver
        self.mapping = mapping
        self.chunksize = definitions.chunk_size
        self.step = 0
        self.collection = project_data['project_name']
        self.db = project_data['db_name']
        self._last_round = False
        # self.completed_jobs = dict.fromkeys(list(self.mapping.keys()), 0)
        self.verbose = 0
        self.n_vars = 0
        self.mongod = mongod_cmd

    def get_sample_csv_vcf_tuple(self):
        """ Locate files associated with a specific sample """

        list_tupls = []

        for _map in self.mapping:

            matching_csv = [i for i in os.listdir(_map['csv_file_full_path']) if i.startswith(_map['csv_file_basename'])
                            and i.endswith('txt')]

            matching_vcf = [i for i in os.listdir(_map['csv_file_full_path']) if i.startswith(_map['csv_file_basename'])
                            and i.endswith('vcf')]

            if len(matching_csv) > 1 or len(matching_vcf) > 1:
                raise ValueError('Too many matching csvs')
            elif len(matching_csv) == 0 or len(matching_vcf) == 0:
                raise ValueError('Csv not found')
            else:
                csv_path = os.path.join(_map['csv_file_full_path'], matching_csv[0])
                vcf_path = os.path.join(_map['csv_file_full_path'], matching_vcf[0])
                list_tupls.append((_map['sample_names'],
                                   vcf_path,
                                   csv_path,
                                   self.db,
                                   self.collection,
                                   _map['extra_data']))

        return list_tupls

    def parallel_annotation(self, n_processes, verbose=1):

        """
        Set up variant parsing scheme. Since a functional programming style is required for parallel
        processing, the input to the Pool.map function from the multiprocessing library must be
        immutable. We use tuples of the type:

            (samples, vcf_path, csv_path, db_name, collection_name, step)

        The parallel annotation will split the vcf, csv file in N steps where N equals the number of lines
        (variants) to be annotated divided a pre-defined 'chunk size'. For instance, a vcf, csv pair containing
        250 000 variants and a chunksize of 5 000 will yield 50 steps. Given 8 parallel procrsses, the annotation
        will happen over those 50 steps with 8 cores working cuncurrently until the 50 jobs are fully annotated.

        :param n_processes: number of cores to be used
        :param verbose: verbosity level [0,1,2,3]
        :return: None
        """

        self.verbose = verbose
        list_tupls = self.get_sample_csv_vcf_tuple()
        map_job = []
        for tpl in list_tupls:
            hgvs = HgvsParser(tpl[1])
            csv_parsing = TxtParser(tpl[2], samples=hgvs.samples, extra_data=tpl[5])
            num_lines = csv_parsing.num_lines
            n_steps = int(num_lines/self.chunksize) + 1
            map_job.extend(self.parallel_annotator_mapper(tpl, n_steps, extra_data=tpl[5], mongod_cmd=self.mongod))

        pool = Pool(n_processes)
        for _ in tqdm.tqdm(pool.imap_unordered(parse_by_step, map_job), total=len(map_job)):
            pass
        pool.close()
        pool.join()
        # logger.info('Completed annotation and parsing for variants in sample %s' % tpl[0])

    @staticmethod
    def parallel_annotator_mapper(_tuple, n_steps, extra_data=None, mongod_cmd=None):
        """ Assign step number to each tuple to be consumed by parsing function """
        new_tuple_list = []
        if mongod_cmd:
            mongod = mongod_cmd
        else:
            mongod = None

        for i in range(n_steps):
            sample = _tuple[0]
            vcf_file = _tuple[1]
            csv_file = _tuple[2]
            db_name = _tuple[3]
            collection_name = _tuple[4]
            extra_data = extra_data
            step = i
            new_tuple_list.append((sample,
                                   vcf_file,
                                   csv_file,
                                   extra_data,
                                   db_name,
                                   collection_name,
                                   step,
                                   mongod))
        return new_tuple_list

    def quick_annotate_and_save(self, n_processes=8):
        """ Annotation that doesn't require annovar """

        list_tuples = []
        for _map in self.mapping:
            simple_map = {k: v for k, v in _map.items() if k not in ['csv_file_basename', 'sample_names']}
            list_tuples.append((simple_map['raw_vcf_file_full_path'], self.db, self.collection))

        for _tuple in list_tuples:

            hgvs = HgvsParser(_tuple[0])
            num_lines = hgvs.get_num_lines()
            n_steps = int(num_lines / self.chunksize) + 1
            map_job = self.quick_annotate_mapper(_tuple, n_steps)
            pool = Pool(n_processes)
            for _ in tqdm.tqdm(pool.imap_unordered(parallel_get_dict_mv, map_job), total=len(map_job)):
                pass

    @staticmethod
    def quick_annotate_mapper(_tuple, n_steps):
        new_tuple_list = []
        for i in range(n_steps):
            step = i
            vcf_path = _tuple[0]
            db_name = _tuple[1]
            collec_name = _tuple[2]
            new_tuple_list.append((vcf_path,
                                   db_name,
                                   collec_name,
                                   step))
        return new_tuple_list

    def generate_output_files_by_sample(self):

        client = MongoClient()
        db = getattr(client, self.project_data['db_name'])
        collection = getattr(db, self.project_data['collection_name'])

        # Get all distinct sample_ids
        samples = collection.distinct('sample_id')

        fwriter = Writer(self.project_data['db_name'], self.project_data['collection_name'])
        filt = Filters(self.project_data['db_name'], self.project_data['collection_name'])

        # Generate files for each sample
        for sample in samples:
            q = filt.variants_from_sample(sample)
            list_docs = list(q)
            out_path_by_sample = os.path.join(self.output_csv_path, 'csv_by_sample')
            if not os.path.exists(out_path_by_sample):
                os.makedirs(out_path_by_sample)
            fname = os.path.join(out_path_by_sample, 'annotated_csv_' + sample + '_all_vars.csv')
            fwriter.generate_annotated_csv(list_docs, fname)


def parse_by_step(maps):
    """ The function that implements the parsing """

    sample = maps[0]
    vcf_file = maps[1]
    csv_file = maps[2]
    extra_data = maps[3]
    db_name = maps[4]
    collection_name = maps[5]
    step = maps[6]
    mongod_cmd = maps[7]

    client = MongoClient(maxPoolSize=None, waitQueueTimeoutMS=200)
    db = getattr(client, db_name)
    collection = getattr(db, collection_name)
    hgvs = HgvsParser(vcf_file)
    csv_parsing = TxtParser(csv_file, samples=sample, extra_data=extra_data)
    list_hgvs_ids = hgvs.get_variants_from_vcf(step)
    myvariants_variants = get_dict_myvariant(list_hgvs_ids, 1, sample)

    csv_variants = csv_parsing.open_and_parse_chunks(step, build_ver='hg19')

    merged_list = []
    for i, _ in enumerate(myvariants_variants):
        for dict_from_sample in csv_variants[i]:
            merged_list.append(merge_dict_lists(myvariants_variants[i], dict_from_sample))

    return insert_handler(merged_list, collection, mongod_cmd)


def insert_handler(merged_list, collection, mongod_cmd):

    logging.info('Parsing Buffer...')
    if len(merged_list) == 0:
        logging.info('Empty list of documents trying to be parsed, skipping and continuing operation...')
        return None
    else:
        try:
            collection.insert_many(merged_list, ordered=False)
        except Exception, error:
            if "Connection refused" in str(error):
                if mongod_cmd:
                    logging.info('MongoDB Server seems to be off. Attempting restart...')
                    print(mongod_cmd)
                    args = shlex.split(mongod_cmd)
                    subprocess.Popen(args, stdout=subprocess.PIPE)
                else:
                    logging.error('Error connecting to mongodb.' + str(error))
                logging.info('Retrying to parse...')
                time.sleep(4)
                insert_handler(merged_list, collection, mongod_cmd)  # Recurse!
            else:
                logging.error('Unrecoverable error: ' + str(error))

    return None


def merge_dict_lists(*dict_args):
    """
    Given any number of dicts, shallow copy and merge into a new dict,
    precedence goes to key value pairs in latter dicts.
    """
    result = {}
    for dictionary in dict_args:
        result.update(dictionary)
    return result


def parallel_get_dict_mv(maps):

    vcf_path = maps[0]
    db_name = maps[1]
    collection_name = maps[2]
    step = maps[3]

    client = MongoClient(maxPoolSize=None, waitQueueTimeoutMS=200)
    db = getattr(client, db_name)
    collection = getattr(db, collection_name)
    hgvs = HgvsParser(vcf_path)

    list_hgvs_ids = hgvs.get_variants_from_vcf(step)
    myvariants_variants = get_dict_myvariant(list_hgvs_ids, 1, hgvs.samples)

    # logging.info('Parsing Buffer...')
    collection.insert_many(myvariants_variants, ordered=False)
    client.close()
    return None


def get_dict_myvariant(variant_list, verbose, sample_id):
    """ Retrieve variants from MyVariant.info"""

    if verbose >= 2:
        verbose = True
    else:
        verbose = False

    mv = myvariant.MyVariantInfo()
    # This will retrieve a list of dictionaries
    try:
        variant_data = mv.getvariants(variant_list, verbose=1, as_dataframe=False)
    except Exception, error:
        logging.info('Error: ' + str(error) + 'while fetching from MyVariant, retrying...')
        time.sleep(5)
        variant_data = get_dict_myvariant(variant_list, verbose, sample_id)

    variant_data = remove_id_key(variant_data, sample_id)
    return variant_data


def remove_id_key(variant_data, sample_id):
    """ Let mongo create an _id key to prevent insert attempts of documents with same key """

    for dic in variant_data:
        dic['hgvs_id'] = dic.pop("_id", None)
        dic['hgvs_id'] = dic.pop("query", None)
        dic['sample_id'] = sample_id

    return variant_data






#### DEPRECATED #####

import VAPr.definitions as definitions
from VAPr.models import TxtParser, HgvsParser
import logging
logger = logging.getLogger()
logger.setLevel(logging.INFO)
try:
    logger.handlers[0].stream = sys.stdout
except:
    pass


class VariantParsing:

    def __init__(self,
                 input_dir,
                 output_csv_path,
                 annovar_path,
                 project_data,
                 mapping,
                 design_file=None,
                 build_ver=None):

        """ Project data """
        self.input_dir = input_dir
        self.output_csv_path = output_csv_path
        self.annovar = annovar_path
        self.project_data = project_data
        self.design_file = design_file
        self.buildver = build_ver
        self.mapping = mapping

        self.chunksize = definitions.chunk_size
        self.step = 0
        self.collection = project_data['project_name']
        self.db = project_data['db_name']
        self._last_round = False
        # self.completed_jobs = dict.fromkeys(list(self.mapping.keys()), 0)
        self.verbose = 0
        self.mongo_client = self.mongo_client_setup()
        self.n_vars = 0


    @staticmethod
    def merge_dict_lists(*dict_args):
        """
        Given any number of dicts, shallow copy and merge into a new dict,
        precedence goes to key value pairs in latter dicts.
        """
        result = {}
        for dictionary in dict_args:
            result.update(dictionary)
        return result


    def get_dict_myvariant(self, variant_list, verbose, sample_id):
        """
        Function designated to place the queries on myvariant.info servers.

        :param variant_list: list of HGVS variant ID's. Usually retrived beforehand using the method
        get_variants_from_vcf from the class VariantParsing.
        :param verbose:
        :return: list of dictionaries. Each dictionary contains data about a single variant.
        """

        if verbose >= 2:
            verbose = True
        else:
            verbose = False

        mv = myvariant.MyVariantInfo()
        # This will retrieve a list of dictionaries
        variant_data = mv.getvariants(variant_list, verbose=verbose, as_dataframe=False)
        variant_data = self.remove_id_key(variant_data, sample_id)

        return variant_data

    def mongo_client_setup(self):
        """
        Setup MongoDB client
        :return: null
        """
        client = MongoClient(maxPoolSize=None, waitQueueTimeoutMS=200)
        db = getattr(client, self.db)
        collection = getattr(db, self.collection)

        return collection

    @staticmethod
    def remove_id_key(variant_data, sample_id):
        for dic in variant_data:
            dic['hgvs_id'] = dic.pop("_id", None)
            dic['hgvs_id'] = dic.pop("query", None)
            dic['sample_id'] = sample_id

        return variant_data


    def annotate_and_saving(self, buffer_vars=False, verbose=2):
        self.verbose = verbose

        list_tupls = self.get_sample_csv_vcf_tuple()
        print(list_tupls)

        for tpl in list_tupls:
            hgvs = HgvsParser(tpl[1])
            test = hgvs.get_variants_from_vcf(1)
            print(len(test))
            csv_parsing = TxtParser(tpl[2], samples=hgvs.samples)
            variant_buffer = []
            self.step = 0
            num_parsed = 0

            while csv_parsing.num_lines > self.step * self.chunksize:

                print('STEP: %i, CHNKSIZE: %i, MULT: %i' % (self.step, self.chunksize, self.step * self.chunksize))
                list_hgvs_ids = hgvs.get_variants_from_vcf(self.step)
                all_ids = hgvs.get_all_variants_from_vcf()
                print('Len List HGVC ID: %i, ALL IDS: %i' % (len(list_hgvs_ids), len(all_ids)),
                      'Num Lines: %i' % csv_parsing.num_lines)

                myvariants_variants = self.get_dict_myvariant(list_hgvs_ids, verbose, tpl[0])
                num_parsed += len(list_hgvs_ids)

                offset = len(list_hgvs_ids) - self.chunksize
                csv_variants = csv_parsing.open_and_parse_chunks(self.step, build_ver=self.buildver,
                                                                 offset=offset)

                print('CSV_VARIANTS: %i' % len(csv_variants))
                if self.verbose >= 1:
                    logger.info('Gathered %i variants so far for sample %s, vcf file %s' % (num_parsed,
                                                                                            tpl[0],
                                                                                            tpl[1]))
                merged_list = []
                for i, _ in enumerate(myvariants_variants):
                    for dict_from_sample in csv_variants[i]:
                        merged_list.append(self.merge_dict_lists(myvariants_variants[i], dict_from_sample))

                variant_buffer.extend(merged_list)
                print('VARBUFFLEN: %i' % len(variant_buffer))
                logging.info('Parsing Buffer...')
                if len(variant_buffer) == 0:
                    print('VarBuff Len == 0, no clue why')
                else:
                    self.collection.insert_many(variant_buffer, ordered=False)
                variant_buffer = []
                self.step += 1

                if len(list_hgvs_ids) < self.chunksize:
                    self._last_round = True

                if self._last_round:  # or (len(variant_buffer) > self._buffer_len):
                    logging.info('Done parsing variants for file pair %s, %s' % (tpl[1], tpl[2]))
                    return 'Done'

        return 'Done'



    def parallel_annotation_efficient(self, n_processes, verbose=1):
        self.verbose = verbose
        samples = self.mapping.keys()
        process_mapping = dict.fromkeys(['process_%i' % i for i in range(n_processes)], 0)

        for sample in samples:
            list_tupls = self.get_sample_csv_vcf_tuple(sample)
            for tpl in list_tupls:
                hgvs = HgvsParser(tpl[1])
                csv_parsing = TxtParser(tpl[2])

            self.pooling(n_processes, list_tupls)
            logger.info('Completed annotation and parsing for variants in sample %s' % sample)


    def _variant_parsing_efficitent(self, maps):
        hgvs = HgvsParser(maps[1])
        csv_parsing = TxtParser(maps[2])

        variant_buffer = []
        n_vars = 0

        # for i in n_processes:
        #    chunk_n =

        while csv_parsing.num_lines > self.step * self.chunksize:

            list_hgvs_ids = hgvs.get_variants_from_vcf(self.step)
            myvariants_variants = self.get_dict_myvariant(list_hgvs_ids, self.verbose, maps[0])

            offset = len(list_hgvs_ids) - self.chunksize
            csv_variants = csv_parsing.open_and_parse_chunks(self.step, build_ver=self.buildver, offset=offset)

            merged_list = []
            for i, _ in enumerate(myvariants_variants):
                merged_list.append(
                    self.merge_dict_lists(myvariants_variants[i], csv_variants[i][j]) for j in csv_variants[i])

            variant_buffer.extend(merged_list)
            n_vars += len(merged_list)
            if self.verbose >= 1:
                logger.info('Gathered %i variants so far for sample %s, vcf file %s' % (n_vars, maps[0], maps[1]))
            self.step += 1

            if len(merged_list) < self.chunksize:
                self._last_round = True

            if (len(variant_buffer) > self._buffer_len) or self._last_round:
                logging.info('Parsing Buffer...')
                self.export(variant_buffer)
                variant_buffer = []

                if self._last_round:
                    self.completed_jobs[maps[0]] += 1
                    return self.completed_jobs

        return self.completed_jobs


import os
import sys
import shlex
import glob
import csv
import time
import datetime
import subprocess
import logging
from VAPr import definitions

logger = logging.getLogger()
logger.setLevel(logging.INFO)
try:
    logger.handlers[0].stream = sys.stdout     # Enables logging on jupyter notebooks
except:
    pass

__author__ = 'Carlo Mazzaferro<cmazzafe@ucsd.edu>'


class AnnovarWrapper:

    """ Wrapper around Annovar download and annotation functions """

    def __init__(self,
                 input_dir,
                 output_csv_path,
                 annovar_path,
                 project_data,
                 mapping,
                 design_file=None,
                 build_ver=None):

        """ Project data """
        self.input_dir = input_dir
        self.output_csv_path = output_csv_path
        self.annovar = annovar_path
        self.project_data = project_data
        self.design_file = design_file
        self.buildver = build_ver
        self.mapping = mapping
        """ Databases data """
        self.down_dd = definitions.down_dd
        self.annovar_hosted = definitions.annovar_hosted
        self.dl_list_command = definitions.dl_list_command
        self.manual_update = definitions.manual_update
        self.hg_18_databases = definitions.hg_18_databases
        self.hg_19_databases = definitions.hg_19_databases
        self.hg_38_databases = definitions.hg_38_databases
        self.databases = self.get_databases()

    def download_dbs(self, all_dbs=True, dbs=None):
        """
        Implementation of the wrapper around annotate_variation.pl with -downdb as optional arg
        First, it cleans up the humandb/ directory to avoid conflicts, then gets newest versions of databases
        by spawning the jobs using subprocesses

        """

        if len(os.listdir(os.path.join(self.annovar, 'humandb/'))) > 0:
            files = glob.glob(os.path.join(self.annovar, 'humandb/*'))
            for f in files:
                os.remove(f)

        list_commands = self.build_db_dl_command_str(all_dbs, dbs)
        for command in list_commands:
            args = shlex.split(command)
            # Spawn subprocesses
            subprocess.Popen(args, stdout=subprocess.PIPE)
            # run_handler(os.path.join(self.annovar, 'humandb/'), cmds=list_commands, annovar_path=self.annovar)

        return 'Finished downloading databases to {}'.format(os.path.join(self.annovar, 'humandb/'))

    def run_annovar(self, batch_jobs=10, multisample=False):
        """ Spawning Annovar jobs in batches of five files at a time to prevent memory overflow """

        handler = AnnovarJobHandler(batch_jobs, multisample, self.mapping)
        n_files_created = 0
        for index, job in enumerate(handler.chunkenize):
            logging.info('Job %i/%i sent for processing' % (index + 1, len(self.mapping)/batch_jobs + 1))
            n_files_created += len(job)

            for idx, _map in enumerate(job):
                annotation_dir = _map['csv_file_full_path']
                if os.path.isdir(annotation_dir):
                    logging.info('Directory already exists for %s. '
                                 'Writing output files there for file %s.' % (annotation_dir,
                                                                              _map['raw_vcf_file_full_path']))
                else:
                    os.makedirs(annotation_dir)

                vcf_path = _map['raw_vcf_file_full_path']
                csv_path = os.path.join(_map['csv_file_full_path'], _map['csv_file_basename'])
                cmd_string = self.build_annovar_command_str(vcf_path, csv_path, multisample=multisample)
                args = shlex.split(cmd_string)

                subprocess.Popen(args, stdout=subprocess.PIPE)

            logging.info('Annovar jobs submitted for %i files: %s' % (len(job),
                                                                      ', '.join([os.path.basename(
                                                                          i['raw_vcf_file_full_path']) for i in
                                                                                 job])))

            listen(self.output_csv_path, len(job), n_files_created)
            logging.info('Finished running Annovar on this batch')

        logging.info('Finished running Annovar on all files')

    def build_annovar_command_str(self, _vcf, _csv, multisample=False):
        """ Concatenate command string arguments for Annovar jobs """

        # TODO: check for newer version of databases

        dbs = ",".join(list(self.databases.keys()))
        dbs_args = ",".join(list(self.databases.values()))

        if '1000g2015aug' in dbs:
            dbs = dbs.replace('1000g2015aug', '1000g2015aug_all')
        command = " ".join(['perl', os.path.join(self.annovar, 'table_annovar.pl'), _vcf,
                            os.path.join(self.annovar, 'humandb/'), '-buildver', self.buildver, '-out',
                            _csv, '-remove -protocol', dbs,  '-operation',
                            dbs_args, '-nastring .', '-otherinfo -vcfinput'])
        if multisample:
            command += ' -format vcf4 -allsample -withfreq'

        return command

    def build_db_dl_command_str(self, all_dbs, dbs):
        """ Concatenate command string arguments for Annovar download database jobs """

        if not all_dbs:
            for db in dbs:
                if db not in self.databases:
                    raise ValueError('Database %s not supported for build version %s' % (db, self.buildver))
            self.databases = {db: self.databases[db] for db in dbs}

        command_list = []

        for db in self.databases:

            if self.annovar_hosted[db]:
                command_list.append(" ".join(['perl', self.annovar + 'annotate_variation.pl', '-build', self.buildver,
                                              '-downdb', self.down_dd, db, self.annovar + 'humandb/']))
            else:
                command_list.append(" ".join(['perl', self.annovar + 'annotate_variation.pl', '-build', self.buildver,
                                              '-downdb', db, self.annovar + 'humandb/']))
        return command_list

    def check_for_database_updates(self):
        """ Deprecated """

        self.download_dbs(all_dbs=False, dbs=['avdblist'])

        with open(os.path.join(self.annovar, '/humandb' + self.buildver + '_avdblist.txt'), 'r') as db_list:
            reader = csv.reader(db_list, delimiter='\t')
            db_dict = {}
            for i in reader:
                if i[0][-6:] != 'idx.gz':
                    db_dict[i[0][5:]] = [datetime.datetime(int(i[1][0:4]), int(i[1][4:6]), int(i[1][6:8])), i[2]]

        for db in self.manual_update.keys():
            for db_ in db_dict.keys():
                if db_.startswith(db):
                    if db_dict[db_][0] > self.manual_update[db][0]:
                        logging.info('Database %s outdated, will download newer version' % db_)
                        self.download_dbs(all_dbs=False, dbs=[os.path.splitext(os.path.splitext(db_)[0])[0]])

    def get_databases(self):

        if self.buildver == 'hg18':
            databases = self.hg_18_databases
        elif self.buildver == 'hg19':
            databases = self.hg_19_databases
        else:
            databases = self.hg_38_databases

        return databases


def listen(out_path, batch_jobs, n_files):
    """ Little function to check for newly created annotated files """

    added = 0

    while True:
        txts = []
        walker = os.walk(out_path)
        for folder, subfolders, files in walker:
            for _file in files:
                if _file.endswith('txt'):
                    txts.append(_file)

        time.sleep(5)
        new_files = []
        walker = os.walk(out_path)
        for folder, subfolders, files in walker:
            for _file in files:
                if _file.endswith('txt'):
                    new_files.append(_file)

        newly_created = [x for x in new_files if x not in txts]

        if len(newly_created) > 0:
            added += 1
            logging.info('File %i/%i: Annovar finished working on file : ' % (added, n_files) +
                         os.path.basename(newly_created[0]) +
                         '.\n A text file has been created in the %s directory\n' % out_path)
        if added == batch_jobs:
            break
        if len(txts) >= n_files:
            break


class AnnovarJobHandler:

    def __init__(self, batch_jobs, multisample, mapping):

        self.mapping = mapping
        self.multisample = multisample
        self.batch_jobs = batch_jobs
        if self.batch_jobs > len(self.mapping):
            self.batch_jobs = len(self.mapping)/2
        self.chunkenize = self.chunks()

    def chunks(self):
        """Yield successive n-sized chunks from l."""
        for i in range(0, len(self.mapping), self.batch_jobs):
            yield self.mapping[i:i + self.batch_jobs]

    def _next(self):
        return next(self.chunkenize)


if __name__ == '__main__':

    def download_dbs(all_dbs=True, dbs=None):
        """
        Implementation of the wrapper around annotate_variation.pl with -downdb as optional arg
        First, it cleans up the humandb/ directory to avoid conflicts, then gets newest versions of databases
        by spawning the jobs using subprocesses

        """

        if len(os.listdir(os.path.join(annovar, 'humandb/'))) > 0:
            files = glob.glob(os.path.join(annovar, 'humandb/*'))
            for f in files:
                os.remove(f)

        list_commands = self.build_db_dl_command_str(all_dbs, dbs)
        for command in list_commands:
            args = shlex.split(command)


# standard libraries
import logging
import sys
from VAPr import validation

logger = logging.getLogger()
logger.setLevel(logging.INFO)
#try:
#    logger.handlers[0].stream = sys.stdout
#except:
#    pass

GLOBAL_ERROR_COUNT = 0

__author__ = 'Birmingham'

# TODO: Forgive mal-formatted vcf files. Log the error but dont break the annotation.


# TODO: rewrite with lambdas or partials
def ignore_pid(info_value, genotype_info_to_fill):
    return ignore_field(info_value, genotype_info_to_fill, 'PID')


# TODO: rewrite with lambdas or partials
def ignore_pgt(info_value, genotype_info_to_fill):
    return ignore_field(info_value, genotype_info_to_fill, 'PGT')


def ignore_field(info_value, genotype_info_to_fill, subkey):
    return genotype_info_to_fill


def fill_genotype_class(alleles, genotype_info_to_fill):
    genotype_class = "homozygous"
    genotype_subclass = "reference"
    alt_subclass_name = "alt"

    if alleles[0] != alleles[1]:
        genotype_class = "heterozygous"
        alt_subclass_name = "compound"

    if "0" not in alleles:
        genotype_subclass = alt_subclass_name

    result = {genotype_class: genotype_subclass}
    genotype_info_to_fill.genotype_subclass_by_class = result
    return result


def fill_genotype(info_value, genotype_info_to_fill):
    global GLOBAL_ERROR_COUNT
    alleles = info_value.split('/')
    if len(alleles) != 2:
        GLOBAL_ERROR_COUNT += 1
        return genotype_info_to_fill
    genotype_info_to_fill.genotype = info_value
    fill_genotype_class(alleles, genotype_info_to_fill)
    return genotype_info_to_fill


def fill_unfiltered_reads_counts(info_value, genotype_info_to_fill):
    global GLOBAL_ERROR_COUNT
    delimiter = ','
    counts = info_value.split(delimiter)
    if len(counts) < 2:
        GLOBAL_ERROR_COUNT += 1
    for curr_count in counts:
        new_allele = Allele(curr_count)
        genotype_info_to_fill.alleles.append(new_allele)
    return genotype_info_to_fill


def fill_filtered_reads_count(info_value, genotype_info_to_fill):
    genotype_info_to_fill.filter_passing_reads_count = info_value
    return genotype_info_to_fill


def fill_genotype_confidence(info_value, genotype_info_to_fill):
    genotype_info_to_fill.genotype_confidence = info_value
    return genotype_info_to_fill


def fill_genotype_likelihoods(info_value, genotype_info_to_fill):
    global GLOBAL_ERROR_COUNT
    generate_alleles = False
    delimiter = ','
    likelihoods = info_value.split(delimiter)

    num_expected_alleles = len(genotype_info_to_fill.alleles)
    if num_expected_alleles == 0:
        generate_alleles = True
        genotype_info_to_fill.alleles.append(Allele(None))

    allele_number = 0
    likelihood_number = 0
    for index in range(len(likelihoods)):
        if likelihood_number > allele_number:
            if generate_alleles:
                genotype_info_to_fill.alleles.append(Allele(None))
            allele_number += 1
            likelihood_number = 0
            if allele_number >= len(genotype_info_to_fill.alleles):
                GLOBAL_ERROR_COUNT += 1
        new_likelihood = GenotypeLikelihood(likelihood_number, allele_number, likelihoods[index])
        genotype_info_to_fill.genotype_likelihoods.append(new_likelihood)
        likelihood_number += 1

    if allele_number < (num_expected_alleles-1) or likelihood_number < num_expected_alleles:
        GLOBAL_ERROR_COUNT += 1

    return genotype_info_to_fill


class VCFGenotypeInfo:

    def __init__(self, raw_string):
        self.db_id = None
        self.genotype = None
        self._genotype_confidence = None
        self._filter_passing_reads_count = None  # NULL
        self.raw_string = raw_string
        self.alleles = []  # 0 for ref, 1 for first alt, etc
        self.genotype_likelihoods = []
        self.errors = GLOBAL_ERROR_COUNT

    @property
    def is_null_call(self):
        result = True if self.raw_string.startswith('./.:') else False
        return result

    @property
    def genotype_confidence(self):
        return self._genotype_confidence

    @genotype_confidence.setter
    def genotype_confidence(self, value):
        # TODO: Determine if genotype confidence value is limited to being a positive or non-negative number
        self._genotype_confidence = validation.convert_to_nullable(value, float)

    @property
    def filter_passing_reads_count(self):
        return self._filter_passing_reads_count

    @filter_passing_reads_count.setter
    def filter_passing_reads_count(self, value):
        self._filter_passing_reads_count = validation.convert_to_nonneg_int(value, nullable=True)[0]
        self.errors += validation.convert_to_nonneg_int(value, nullable=True)[1]


class Allele:
    def __init__(self, read_counts=None):
        self.db_id = None
        self._read_counts = None
        if read_counts is not None:
            self.read_counts = read_counts
        else:
            self._read_counts = 'NULL'
        self.errors = GLOBAL_ERROR_COUNT

    @property
    def read_counts(self):
        return self._read_counts

    @read_counts.setter
    def read_counts(self, value):
        self._read_counts = validation.convert_to_nonneg_int(value, nullable=True)[0]
        self.errors += validation.convert_to_nonneg_int(value, nullable=True)[1]


class GenotypeLikelihood:
    @staticmethod
    def _validate_allele_relationship(allele1_number, allele2_number):
        global GLOBAL_ERROR_COUNT
        if allele1_number > allele2_number:
            GLOBAL_ERROR_COUNT += 1

    def __init__(self, allele1_number, allele2_number, likelihood_neg_exponent):

        self.db_id = None
        self._allele1_number = None
        self._allele2_number = None
        self._likelihood_neg_exponent = None
        self.allele1_number = allele1_number
        self.allele2_number = allele2_number
        self.likelihood_neg_exponent = likelihood_neg_exponent
        self.genotype_subclass_by_class = None
        self.errors = GLOBAL_ERROR_COUNT

    @property
    def allele1_number(self):
        return self._allele1_number

    @allele1_number.setter
    def allele1_number(self, value):
        int_value = validation.convert_to_nonneg_int(value, nullable=True)[0]
        self.errors += validation.convert_to_nonneg_int(value, nullable=True)[1]
        if self.allele2_number is not None:
            self._validate_allele_relationship(int_value, self.allele2_number)
        self._allele1_number = int_value

    @property
    def allele2_number(self):
        return self._allele2_number

    @allele2_number.setter
    def allele2_number(self, value):
        int_value = validation.convert_to_nonneg_int(value, nullable=True)[0]
        self.errors += validation.convert_to_nonneg_int(value, nullable=True)[1]
        if self.allele1_number is not None:
            self._validate_allele_relationship(self.allele1_number, int_value)
        self._allele2_number = int_value

    @property
    def likelihood_neg_exponent(self):
        return self._likelihood_neg_exponent

    @likelihood_neg_exponent.setter
    def likelihood_neg_exponent(self, value):
        self._likelihood_neg_exponent = validation.convert_to_nullable(value, float)

    # GT:AD:BQ:DP:FA

# AD BQ DP FA GQ GT PL SS

class VCFGenotypeStrings:
    #TODO: Parser for Mutect files

    _DELIMITER = ':'
    _PARSER_FUNCS = {'GT': fill_genotype,
                     'AD': fill_unfiltered_reads_counts,
                     # 'BQ': fill_base_quality,
                     'DP': fill_filtered_reads_count,
                     # 'FA': fill_allele_fraction,
                     'GQ': fill_genotype_confidence,
                     'PL': fill_genotype_likelihoods,
                     # 'SS': fill_variant_status,
                     'PID': ignore_pid,
                     'PGT': ignore_pgt}

    @classmethod
    def parse(cls, format_string, info_string):

        mutect_formats = ['BQ', 'FA', 'SS']
        possible_string_formats = ['GT:GQ:PL', 'GT:AD:GQ:PL', 'GT:AD:DP:GQ:PL', 'GT:AD:DP:GQ:PGT:PID:PL']
        result = VCFGenotypeInfo(info_string)

        if not result.is_null_call:
            info_subkeys = format_string.split(cls._DELIMITER)
            info_values = info_string.split(cls._DELIMITER)

            for index, value in enumerate(info_subkeys):
                if value in mutect_formats:
                    continue
                if value not in possible_string_formats[-1]:
                    continue
                curr_key = info_subkeys[index]
                curr_value = info_values[index]
                parse_func = cls._PARSER_FUNCS[curr_key]
                result = parse_func(curr_value, result)

        return result

"""
if __name__ == '__main__':

    for i in range(0,10000):
        if i%2:
            dictionary = {'otherinfo': ['GT:AD:BQ:DP:FA', '0:0,3:.:3:1.00']}
        else:
            dictionary = {'otherinfo': ['GT:AD:GQ:PL', '1/1:0,43:99:1039,129,0']}
        parser = VCFGenotypeStrings()

        genotype_to_fill = parser.parse(dictionary['otherinfo'][0], dictionary['otherinfo'][1])
        gen_dic = {'genotype': genotype_to_fill.genotype}

        genotype = genotype_to_fill.genotype
        filter_passing_reads_count = genotype_to_fill.filter_passing_reads_count
        genotype_likelihoods = genotype_to_fill.genotype_likelihoods
        alleles = genotype_to_fill.alleles

        try:
            gen_dic['filter_passing_reads_count'] = [int(genotype_to_fill.filter_passing_reads_count)]
        except ValueError:
            print('Read depth returned null value')
        try:
            gen_dic['genotype_likelihoods'] = [float(genotype_to_fill.genotype_likelihoods[0].likelihood_neg_exponent),
                                               float(genotype_to_fill.genotype_likelihoods[1].likelihood_neg_exponent),
                                               float(genotype_to_fill.genotype_likelihoods[2].likelihood_neg_exponent)]
        except IndexError:
            print('Genotype Likelihood not understood')

        try:
            gen_dic['alleles'] = [int(genotype_to_fill.alleles[0].read_counts),
                                  int(genotype_to_fill.alleles[1].read_counts)]
        except IndexError:
            print('Allele read counts  not understood')

        print(gen_dic)
"""


import os
import logging
import vcf
import sys
logger = logging.getLogger()
logger.setLevel(logging.INFO)
try:
    logger.handlers[0].stream = sys.stdout
except:
    pass

__author__ = 'Carlo Mazzaferro<cmazzafe@ucsd.edu>'


class Ingester:

    def __init__(self, input_dir, output_dir):

        self.base_dir = input_dir
        self.out_dir = output_dir
        self.mapping_list = []
        self.atomic_ingester = SingleInstance

    @staticmethod
    def listdir_fullpath(d):
        return [os.path.join(d, f) for f in os.listdir(d)]

    def check_contents(self):
        """ Basic checker for input files """
        types = {'files': [], 'dirs': [], 'other': 0}
        contents = self.listdir_fullpath(self.base_dir)

        for path in contents:
            if os.path.isfile(path):
                types['files'].append(('file', path))
            elif os.path.isdir(path):
                types['dirs'].append(('dir', path))
            else:
                types['other'] += 1

        if len(set(types)) > 1:
            raise IOError('Files and dir')

    def walk(self):
        """ Walks the files, subdirs and dirs in input directory"""
        walker = os.walk(self.base_dir)
        for folder, subfolders, files in walker:
            for _file in files:
                full_path_single_file = os.path.join(os.path.abspath(folder), _file)
                self.digest_single(full_path_single_file)

    def digest_single(self, single, sample='infer', ingester_type='files', extra_data=None):
        """ Digests input data from input directory """

        ingested = self.atomic_ingester(single, self.base_dir, self.out_dir, sample=sample, ingester_type=ingester_type,
                                        extra_data=extra_data)

        self.mapping_list.append(ingested.mapping)

        # Eliminates possible duplicates
        self.mapping_list = list({v['raw_vcf_file_full_path']: v for v in self.mapping_list}.values())

    def digest_design_file(self, design_df):
        """ Digests input data from csv file """

        vcf_files = []
        design_file_mapping = design_df.set_index('Sample_Names').T.to_dict()
        for sample in design_file_mapping.keys():
            if sample.endswith('.vcf'):   # Design file contains file names

                sample_dir = self.base_dir
                vcf_file = [i for i in os.listdir(sample_dir) if i.startswith(sample)]
                if len(vcf_file) > 1:
                    raise NameError('More tha one file found with unique name in design file')
                else:
                    logging.info('Found %i unique vcf files for sample %s' % (len(set(vcf_files)), sample))
                    full_path_single_file = os.path.join(os.path.abspath(sample_dir), vcf_file[0])
                    self.digest_single(full_path_single_file, sample=sample, ingester_type='files',
                                       extra_data=design_file_mapping[sample])

            else:                         # Design file contains directory (sample) names
                sample_dir = os.path.join(self.base_dir, sample)
                if not os.path.exists(sample_dir) or os.listdir(sample_dir) == []:
                    raise NameError('Could not find directory named %s as provided in design file' % sample)

                vcf_files = [i for i in os.listdir(sample_dir) if i.endswith('.vcf')]

                for vcf_file in vcf_files:
                    full_path_single_file = os.path.join(os.path.abspath(sample_dir), vcf_file)
                    self.digest_single(full_path_single_file, sample=sample, ingester_type='dirs',
                                       extra_data=design_file_mapping[sample])


class SingleInstance:

    """ Process data of single vcf file, populates dictionary """

    def __init__(self, single_instance, input_dir, out_dir, sample='infer', ingester_type='files', extra_data=None):

        self.instance = single_instance
        self.out_dir = out_dir
        self.input_dir = input_dir
        self.reader = vcf.Reader(open(single_instance, 'r'))
        self.sample = sample
        self.ingester_type = ingester_type
        self.extra_data = extra_data
        self.mapping = {'raw_vcf_file_full_path': self.fill_input_vcf_path(),
                        'vcf_file_basename': self.fill_vcf_file_basename(),
                        'csv_file_basename': self.fill_csv_file_basename(),
                        'sample_names': self.fill_sample_names(),
                        'num_samples_in_csv': len(self.fill_sample_names()),
                        'csv_file_full_path': self.fill_csv_sample_dir(),
                        'vcf_sample_dir': self.fill_vcf_sample_dir()}
        self.add_extra_data()
        self.create_path()

    def add_extra_data(self):
        if self.extra_data:
            self.mapping['extra_data'] = self.extra_data
        else:
            self.mapping['extra_data'] = None

    def add_key(self, key, value):
        self.mapping[key] = value

    def fill_sample_names(self):

        if self.sample == 'infer':
            return self.reader.samples
        else:
            return [self.sample]

    def sample_dir_name(self):
        if self.ingester_type == 'dirs':
            return self.sample
        else:
            return ''

    def fill_input_vcf_path(self):
        return os.path.abspath(self.instance)

    def fill_vcf_file_basename(self):
        return os.path.basename(self.instance)

    def fill_csv_file_basename(self):
        return os.path.splitext(os.path.basename(self.instance))[0] + '_annotated'

    def fill_csv_sample_dir(self):
        return os.path.join(self.out_dir, self.sample_dir_name())

    def fill_vcf_sample_dir(self):
        return os.path.join(self.input_dir, self.sample_dir_name())

    def create_path(self):
        """ Creates directory named after the sample in a vcf file """
        try:
            os.mkdir(self.mapping['csv_file_full_path'])
        except OSError:
            logging.info('Csv output dir %s for sample exists, moving files there' % self.mapping['csv_file_full_path'])

    def move_file(self):
        """ Moves files to newly created directory """
        try:
            os.rename(self.mapping['raw_vcf_file_full_path'], os.path.join(self.mapping['vcf_sample_dir'],
                                                                           self.mapping['vcf_file_basename']))

            self.mapping['raw_vcf_file_full_path'] = os.path.join(self.mapping['vcf_sample_dir'],
                                                                  self.mapping['vcf_file_basename'])
        except OSError:
            logging.info('Files are organized already')

        print(os.path.isfile(self.mapping['raw_vcf_file_full_path']), self.mapping['raw_vcf_file_full_path'])


from pymongo import MongoClient


class Filters(object):

    """
    Class for variant filtering. DB and collection names can be passed directly to the class and will be used in each
    method.

    Description:

        Rare variants filters are aimed at finding variants that occur with low frequency in the general population.
        Further, some of the filters implemented also target variants whose gene function is known. Higher selectivity
        filters such as the rare_high_impact_variant filter also screens variant data function using CADD Phred scores.

    """

    def __init__(self, db_name, collection_name):

        self.collection_name = collection_name
        self.db_name = db_name

    def variants_from_sample(self, sample_name):

        client = MongoClient()
        db = getattr(client, self.db_name)
        collection = getattr(db, self.collection_name)
        filtered = collection.find({'sample_id': sample_name})
        return list(filtered)

    def rare_cancer_variant(self, samples):
        """ Function for retrieving rare cancer variants """

        client = MongoClient()
        db = getattr(client, self.db_name)
        collection = getattr(db, self.collection_name)
        if not samples:
            samples = collection.distinct('sample_id')

        filtered = collection.find(
            {
                "$and":
                    [
                        {
                            "$or":
                                [
                                    {"esp6500siv2_all": {"$lt": 0.05}},
                                    {"esp6500siv2_all": {"$exists": False}}
                                ]
                        },
                        {
                            "$or":
                                [
                                    {"func_knowngene": "exonic"},
                                    {"func_knowngene": "splicing"}
                                ]
                        },
                        {"exonicfunc_knowngene": {"$ne": "synonymous SNV"}},
                        {"filter_passing_reads_count": {"$gte": 10}},
                        {"cosmic70": {"$exists": True}},
                        {"1000g2015aug_all": {"$lt": 100}},
                        {"sample_id": {"$in": samples}}
                    ]
            }
        )

        filtered = list(filtered)
        print('Variants found that match rarity criteria: {}'.format(len(filtered)))
        return filtered

    def rare_disease_variant(self, samples=None):
        """ Function for retrieving rare disease variants. Includes filter on Clinvar data  """

        client = MongoClient()
        db = getattr(client, self.db_name)
        collection = getattr(db, self.collection_name)
        if not samples:
            samples = collection.distinct('sample_id')

        filtered = collection.find(
            {
                "$and":
                    [
                        {
                            "$or":
                                [
                                   {"esp6500siv2_all": {"$lt": 0.05}},
                                   {"esp6500siv2_all": {"$exists": False}}
                                ]
                        },
                        {
                            "$or":
                                [
                                    {"func_knowngene": "exonic"},
                                    {"func_knowngene": "splicing"}
                                ]
                        },
                        {"exonicfunc_knowngene": {"$ne": "synonymous SNV"}},
                        {"filter_passing_reads_count": {"$gte": 10}},
                        {
                            "$or":
                                [
                                    {"cosmic70": {"$exists": True}},
                                    {"clinvar": {"$exists": True}}
                                ]
                        },
                        {"1000g2015aug_all": {"$lt": 0.1}},
                        {"sample_id": {"$in": samples}},
                    ]
            }
        )

        filtered = list(filtered)
        print ('Variants found that match rarity criteria: {}'.format(len(filtered)))
        return filtered

    def rare_high_impact_variants(self, samples=None):
        """ Function for retrieving rare disease variants. Includes filter on Clinvar as well CADD Phred data  """

        client = MongoClient()

        db = getattr(client, self.db_name)
        collection = getattr(db, self.collection_name)
        if not samples:
            samples = collection.distinct('sample_id')

        filtered = collection.find(
            {
                "$and":
                    [
                        {
                            "$or":
                                [
                                    {"esp6500siv2_all": {"$lt": 0.05}},
                                    {"esp6500siv2_all": {"$exists": False}}
                                ]
                        },
                        {
                            "$or":
                                [
                                    {"func_knowngene": "exonic"},
                                    {"func_knowngene": "splicing"}
                                ]
                        },
                        {"exonicfunc_knowngene": {"$ne": "synonymous SNV"}},
                        {"filter_passing_reads_count": {"$gte": 10}},
                        {
                            "$or":
                                [
                                    {"cosmic70": {"$exists": True}},
                                    {"clinvar": {"$exists": True}}
                                ]
                        },
                        {"1000g2015aug_all": {"$lt": 0.1}},
                        {"cadd.phred": {"$gte": 15}},
                        {"sample_id": {"$in": samples}},
                    ]
            }
        )

        filtered = list(filtered)
        print('Variants found that match rarity criteria: {}'.format(len(filtered)))
        return filtered

    def deleterious_compound_heterozygote(self, samples=None):
        """ Function for retrieving deleterious compound heterozygote variants  """

        client = MongoClient()

        db = getattr(client, self.db_name)
        collection = getattr(db, self.collection_name)
        if not samples:
            samples = collection.distinct('sample_id')

        filtered = collection.find(
            {
                "$and":
                    [
                        {"genotype_subclass_by_class.heterozygous": "compound"},
                        {"cadd.phred": {"$gte": 10}},
                        {"sample_id": {"$in": samples}},
                    ]
            }
        )

        filtered = list(filtered)
        print('Variants found that match rarity criteria: {}'.format(len(filtered)))
        return filtered

    def get_de_novo(self, sample1, sample2, sample3, multisample=True):
        """
        Function for de novo variant analysis. Can be performed on multisample files or or on data coming
        from a collection of files. In the former case, every sample contains the same variants, although they have
        differences in their allele frequency and read values. A de novo variant is defined as a variant that
        occurs only in the specified sample (sample1) and not on the other two (sample2, sample3). Occurrence is
        defined as having allele frequencies greater than [0, 0] ([REF, ALT]).

        For the case in which the variants come from multiple files, de novo variants are the ones that occur only
        once on the sample specified (usually found in one file), and that do not occur on the other samples (found
        in the other files).
        """

        client = MongoClient()
        db = getattr(client, self.db_name)
        collection = getattr(db, self.collection_name)
        if multisample:

            filtered_hgvs = collection.distinct("hgvs_id",
                                                {
                                                    "$or": [
                                                            {"$and": [
                                                                {"alleles": [0, 0]},
                                                                {"sample_id": sample2}
                                                            ]
                                                            },
                                                            {"$and": [
                                                                {"alleles": [0, 0]},
                                                                {"sample_id": sample3}
                                                            ]
                                                            },
                                                        ]
                                                }
                                                )

            de_novo = collection.find(
                {
                    "$and":
                        [
                            {"hgvs_id": {"$in": filtered_hgvs}},
                            {"alleles": {"$ne": [0, 0]}},
                            {"sample_id": sample1}
                        ]

                }
            )
        else:
            filtered_hgvs = collection.aggregate(
                [
                    {"$group": {
                        "_id": "$hgvs_id",
                        "samples":
                            {"$addToSet": {
                                "$cond": [  # works with a if ... then ... else syntax

                                    {  # if these are satisfied,
                                        "$or": [
                                            {"$eq": ["$sample_id", sample1]},
                                            {"$eq": ["$sample_id", sample2]},
                                            {"$eq": ["$sample_id", sample3]}
                                        ]
                                    },
                                    # then add sample_id to the set of variants
                                    "$sample_id",
                                    # else, add zero
                                    0
                                ]

                            }
                            }
                    }
                    },
                    {"$redact": {
                        "$cond": {
                            "if": {
                                "$and": [
                                    {"$in": [sample1, "$samples"]},
                                    {"$eq": [{"$size": "$samples"}, 1]}  # Occurs only once, hence a unique HGVS_ID
                                ]
                            },
                            "then": "$$KEEP",
                            "else": "$$PRUNE"
                        }
                    }
                    }
                ]
            )
            as_hgvs_list = [i['_id'] for i in filtered_hgvs]
            de_novo = collection.find({'hgvs_id': {'$in': as_hgvs_list}})

        print('Variants found that match de novo criteria: {}'.format(len(de_novo)))
        return list(de_novo)
